#ifndef CURVED_FUNCTION_HH
#define CURVED_FUNCTION_HH

#include <cmath>
#include <dune/fufem/interval.hh>

namespace Dune {
template <class NonlinearityType> class CurvedFunction {
  typedef typename NonlinearityType::VectorType VectorType;
  typedef typename NonlinearityType::MatrixType MatrixType;

public:
  CurvedFunction(MatrixType const &A, VectorType const &b,
                 NonlinearityType const &phi, VectorType const &x,
                 VectorType const &dir)
      : A(A), b(b), phi(phi), x(x), dir(dir) {}

  double quadraticPart() const { return 0; }

  double linearPart() const { return 0; }

  void subDiff(double m, Interval<double> &D) const {
    VectorType x;
    cartesian(m, x);
    VectorType t;
    tangent(m, t);

    phi.directionalSubDiff(x, t, D);

    VectorType tmp;
    A.mv(x, tmp);                //  Ax
    tmp -= b;                    //  Ax - b
    double const dotp = tmp * t; // <Ax - b,t>
    D[0] += dotp;
    D[1] += dotp;
  }

  void domain(Interval<double> &domain) const {
    domain[0] = -M_PI;
    domain[1] = M_PI;
  }

  void cartesian(double m, VectorType &y) const {
    y = 0;
    y.axpy(std::cos(m), x);
    y.axpy(std::sin(m), dir);
  }

private:
  MatrixType const &A;
  VectorType const &b;
  NonlinearityType const &phi;
  VectorType const &x;
  VectorType const &dir;

  /* If x and d were normalised, we would have

       cartesian(a) =  cos(a) * x + sin(a) * d and
       tangent(a)   = -sin(a) * x + cos(a) * d.

     Since we x and d are not normalised and the return of
     cartesian() is fixed, we scale the tangent.
  */
  void tangent(double m, VectorType &y) const {
    y = 0;
    y.axpy(-std::sin(m) * dir.two_norm2(), x);
    y.axpy(std::cos(m) * x.two_norm2(), dir);
  }
};
}

#endif
